#include <algorithm>
#include <array>
#include <chrono>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <locale>
#include <numeric>
#include <optional>
#include <string>
#include <string_view>
#include <sstream>
#include <type_traits>
#include <vector>

#include <cmath>
#include <cstdint>
#include <cstdlib>

extern "C" {
#include <assert.h>
#include <unistd.h>

#include "mpi.h"
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wdeprecated-copy-with-dtor"
  #include "cxxopts.hpp"
#pragma GCC diagnostic pop


// wrapper checking MPI calls for success
#define MPI_ASSERT(call)                                                       \
do {                                                                           \
	int ret = call;                                                             \
	if (MPI_SUCCESS != ret) {                                                   \
		std::cerr << "MPI call " << std::quoted(#call) << " failed with status " \
		  << ret << " (line " << __LINE__ << ")" << std::endl;                   \
		die();                                                                   \
	}                                                                           \
} while (0)


using std::chrono::steady_clock,
      std::chrono::microseconds,
      std::chrono::nanoseconds,
      std::to_string;


namespace {

// why isn't this part of the STL!?
template <typename Rep, typename Ratio>
std::string
get_unit(std::chrono::duration<Rep,Ratio> const &) = delete;

template <typename Rep>
std::string
get_unit(std::chrono::duration<Rep,std::milli> const &)
{
	return "ms";
}

template <typename Rep>
std::string
get_unit(std::chrono::duration<Rep,std::micro> const &)
{
	return "us";
}

template <typename Rep>
std::string
get_unit(std::chrono::duration<Rep,std::nano> const &)
{
	return "ns";
}


template <typename Rep, typename Ratio>
std::string
to_string(std::chrono::duration<Rep,Ratio> const &dur)
{
	std::stringstream ss;

	ss.imbue(std::locale("")); // use locale from environment
	ss << dur.count() << ' ' << get_unit(dur);
	return ss.str();
}

template <typename Rep, typename Ratio>
std::ostream&
operator<<(std::ostream &os, std::chrono::duration<Rep,Ratio> const &dur)
{
	return os << to_string(dur);
}


// Programm configuration
class Config
{
	size_t _iterations,
	       _warmup,
	       _msg_size;
	microseconds _busy_before,
	             _busy_inside,
	             _busy_after;
	std::string _outfile;

public:
	Config() : _iterations{10'000},
	           _warmup{_iterations / 10},
	           _msg_size{128},
	           _busy_before{0},
	           _busy_inside{0},
	           _busy_after{0},
	           _outfile{""} {}

	size_t iterations() const { return _iterations; }
	void iterations(size_t i)
	{
		_iterations = i;
		warmup_iterations(i / 10);
	}

	size_t warmup_iterations() const { return _warmup; }
	void warmup_iterations(size_t i) { _warmup = i; }

	size_t msg_size() const { return _msg_size; }
	void msg_size(size_t val) { _msg_size = val; }

	microseconds busy_before() const { return _busy_before; }
	void busy_before(microseconds val) { _busy_before = val; }

	microseconds busy_inside() const { return _busy_inside; }
	void busy_inside(microseconds val) { _busy_inside = val; }

	microseconds busy_after() const { return _busy_after; }
	void busy_after(microseconds val) { _busy_after = val; }

	std::string outfile() const { return _outfile; }
	void outfile(std::string const val) { _outfile = val; }
};

// Read environment variable
// from: https://stackoverflow.com/a/5401220
[[nodiscard]]
std::optional<std::string>
get_env_var(std::string const &var)
{
	char const *const val = std::getenv(var.c_str());
	return (val != nullptr) ? std::optional<std::string>{val} : std::nullopt;
}

[[nodiscard]]
std::optional<size_t>
get_env_num(std::string const &var)
{
	std::optional<std::string> const val {get_env_var(var)};

	if (! val) { return std::nullopt; }

	char *str_end;
	long long const num{std::strtoll(val->c_str(), &str_end, 0)};

	// sanity checking
	if (num < 0 || *str_end)
	{
		std::cerr << "Use a positive integer value for option "
		  << std::quoted(var) << std::endl;
		exit(EXIT_FAILURE);
	}

	if (static_cast<unsigned long long>(num) > SIZE_MAX)
	{
		std::cerr << "Value too large for option " << std::quoted(var)
		  << ": " << num << std::endl;
		exit(EXIT_FAILURE);
	}

	return std::optional<size_t>{num};
}


// Parse command line options and return a \ref{Config}
[[nodiscard]]
Config
parse_options(int const argc, char const *const *const argv)
try {
	assert(argc >= 1 && "Broken OS not passing programme name");
	Config conf;
	std::string const env_iter{"LIBRA_ITER"},
	                  env_warmup{"LIBRA_WARMUP"},
	                  env_msgsize{"LIBRA_MSGSIZE"};
	std::stringstream help_msg;

	// build our help message
	help_msg << "Libra --- a flexible MPI latency micro benchmark" << std::endl
	  << std::endl
	  << "Runs a configurable number of iterations. In each iteration, rank 0" << std::endl
	  << "will 'MPI_Send' a message of configurable size to rank 1, then 'MPI_Recv'" << std::endl
	  << "a reply. This reply does not contain data (size 0). At rank 0, we measure" << std::endl
	  << "the time from directly before 'MPI_Send' to directly after 'MPI_Recv'." << std::endl
	  << std::endl
	  << "[B] Before, [A] after, and [M] within (between send and receive) each" << std::endl
	  << "ping-pong, the benchmark runs a tight busy loop for a configurable time." << std::endl
	  << "This can simulate imbalance or allow the receiver to enter blocking mode. " << std::endl
	  << "Note that these waiting times might show up in the measured time. This is " << std::endl
	  << "the case for [M] at rank 0 as well as [A] and [M] at rank 1." << std::endl
	  << std::endl
	  << "The runtime for each individual ping-pong can be stored in a file for" << std::endl
	  << "further analysis." << std::endl
	  << std::endl
	  << "Make sure that the values for the number of (warm-up) iterations and the" << std::endl
	  << "message size is *identical* for both processes; consider using environment" << std::endl
	  << "variables (" << env_iter << ", " << env_warmup << ", " << env_msgsize << ")" << std::endl
	  << "to set those values consistently. Note, that parameters always take precedence" << std::endl
	  << "over an environment variable setting the same option." << std::endl;


	cxxopts::Options opts(argv[0], help_msg.str().c_str());

	opts.add_options()
	 ("a,busy_after",
	  "Time (in " + get_unit(conf.busy_after()) + ") to run the busy loop after "
	  "data transfer/between communications, default: " +
	  to_string(conf.busy_after()),
	  cxxopts::value<size_t>())

	 ("b,busy_before",
	  "Time (in " + get_unit(conf.busy_before()) + ") to run the busy loop "
	  "before data transfer, default: " +
	  to_string(conf.busy_before()),
	  cxxopts::value<size_t>())

	 ("m,busy_inside",
	  "Time (in " + get_unit(conf.busy_inside()) + ") to run the busy loop "
	  "during data transfer (between send/receive or vice versa), default: " +
	  to_string(conf.busy_inside()),
	  cxxopts::value<size_t>())

	 ("i,iterations",
	  "Number of iterations/repititions, default: "
	  + std::to_string(conf.iterations()),
	  cxxopts::value<size_t>())

	 ("w,warmup_iterations",
	  "Number of warm-up iterations before measurement, default: "
	  "'iterations / 10'",
	  cxxopts::value<size_t>())

	 ("s,message_size",
	  "Size (in byte) of messages sent, default: "
	  + std::to_string(conf.msg_size()) + " byte",
	  cxxopts::value<size_t>())

	 ("o,outfile", "Output file for verbose stats (no default)",
	  cxxopts::value<std::string>())
	 ("h,help", "Show this help text")
	;

	cxxopts::ParseResult const res = opts.parse(argc, argv);

	// early exit if the user just wants the help message
	if (res.count("help"))
	{
		std::cerr << opts.help() << std::endl;
		exit(EXIT_SUCCESS);
	}

	// try the environment first
	if (auto val{get_env_num(env_iter)}; val) { conf.iterations(*val); }
	if (auto val{get_env_num(env_warmup)}; val) { conf.warmup_iterations(*val); }
	if (auto val{get_env_num(env_msgsize)}; val) { conf.msg_size(*val); }

	// check the command line options
	if (res.count("a")) { conf.busy_after(microseconds{res["a"].as<size_t>()}); }
	if (res.count("b")) { conf.busy_before(microseconds{res["b"].as<size_t>()}); }
	if (res.count("m")) { conf.busy_inside(microseconds{res["m"].as<size_t>()}); }
	if (res.count("i")) { conf.iterations(res["i"].as<size_t>()); }
	if (res.count("w")) { conf.warmup_iterations(res["w"].as<size_t>()); }
	if (res.count("s")) { conf.msg_size(res["s"].as<size_t>()); }
	if (res.count("o")) { conf.outfile(res["o"].as<std::string>()); }

	return conf;
}
catch (cxxopts::OptionException const &e)
{
	std::cerr << "Error parsing options: " << e.what() << std::endl
	  << std::endl
	  << "Try '" << argv[0] << " -h'" << std::endl;
	exit(EXIT_FAILURE);
}


// Simple statistics on our measurements
//
// Takes a 'std::vector' of measurements in 'std::nanoseconds''
// Returns various statistics in 'double' nanoseconds
struct Statistics
{
	nanoseconds sum,
	            mean,
	            median,
	            variance,
	            stddev,
	            min,
	            max;

public:
	Statistics(std::vector<nanoseconds> values)
	{
		size_t const size { values.size() };
		if (! size)
		{
			throw std::length_error("Statistics impossible for empty vectors");
		}

		std::sort(values.begin(), values.end());
		sum  = std::accumulate(values.cbegin(), values.cend(), nanoseconds{0});
		mean = sum / size;
		min  = values[0];
		max  = values[size - 1];
		median = values[size / 2];

		// for anything beyond addition, we ned to mess with type casts

		// from: https://stackoverflow.com/a/48578852
		auto var_accu = [&mean = std::as_const(mean)](nanoseconds accu, nanoseconds const &val) {
			auto const diff{(val - mean).count()};
			return accu + nanoseconds{diff * diff};
		};

		// yes, we devide be 'size - 1' ("Bessel's correction")
		variance = std::accumulate(values.cbegin(), values.cend(), nanoseconds{0}, var_accu) / (size - 1);
		stddev = nanoseconds{static_cast<nanoseconds::rep>(std::sqrt(variance.count()))};
	}
};


// Write detailed timing information to a file, one value per line
bool
write_timings(std::string filename, std::vector<std::chrono::nanoseconds> const &values)
{
	std::ofstream file{filename, std::ios::trunc};

	if (! file.is_open()) { return false; }

	file << "# individual libra measurements; one iteration per line; "
	  << "all values in "
	  << get_unit(std::decay<decltype(values)>::type::value_type{})
	  << std::endl;
	for (nanoseconds const &val : values)
	{
		file << val.count() << std::endl;
		if (!file.good()) {return false; }
	}

	file.close();
	return file.good();
}


// Run a tight loop for (at least) `time` ms
[[maybe_unused]]
inline
void
timed_loop(microseconds dur)
{
	std::chrono::time_point const target { dur + steady_clock::now() };

	while (steady_clock::now() < target) {}
}

// Run a tight ASM loop for `count` iterations
[[maybe_unused]]
inline
void
loop(size_t count)
{
	__asm__ __volatile__ (
		"test %0, %0\n\t" // extra test at the start so we can cope with 'count' == 0
		"loop_start_%=:\n\t" // create a unique "local" label as we will be inlined
		"jz loop_end_%=\n\t"
		"dec %0\n\t"
		"jmp loop_start_%=\n\t"
		"loop_end_%=:\n\t"
		: "=r" (count) // out, fake but needed as we clobber the value
		: "0" (count)  // in
		: "cc"         // clobber: we use the FLAGS register
	);
}

[[noreturn]]
inline
void
die()
{
	MPI_Abort(MPI_COMM_WORLD, EXIT_FAILURE); \
	std::exit(EXIT_FAILURE);
}

} // end -- anon namespace


int
main(int argc, char ** argv)
{
	Config const conf = parse_options(argc, argv);
	std::cout.imbue(std::locale("")); // use locale from environment

	int mpi_rank{0},
	    mpi_size{0};
	MPI_ASSERT(MPI_Init(&argc, &argv));
	MPI_ASSERT(MPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank));
	MPI_ASSERT(MPI_Comm_size(MPI_COMM_WORLD, &mpi_size));

	if (mpi_size != 2)
	{
		std::cerr << "Exactly 2 ranks required" << std::endl;
		die();
	}

	// output our parameters, prefixed by our rank
	{
		// get our node's name just for this output
		char mpi_name[MPI_MAX_PROCESSOR_NAME];
		int mpi_len;
		MPI_ASSERT(MPI_Get_processor_name(mpi_name, &mpi_len));

		// collect everything first to avoid tearing
		std::stringstream ss_id, ss;
		ss.imbue(std::locale("")); // use locale from environment

		ss_id << "[" << mpi_rank << "] ";
		std::string const id{ss_id.str()};

		ss << id << "Host: " << std::string_view{mpi_name, static_cast<unsigned>(mpi_len)} << std::endl
		   << id << "Iterations: " << conf.iterations()  << std::endl
		   << id << "Warm-up: " << conf.warmup_iterations() << std::endl
		   << id << "Busy time before comm: " << conf.busy_before() << std::endl
		   << id << "Busy time inside comm: " << conf.busy_inside() << std::endl
		   << id << "Busy time after comm: " << conf.busy_after() << std::endl
		   << id << "Message size: " << conf.msg_size() << " byte" << std::endl
		   << std::endl;
		std::cout << ss.str();
	}

	// ========================================================================//
	// measure
	steady_clock::time_point time_start{},
	                         time_end{steady_clock::now()};

	// detailed measurement data and message buffer
	//
	// pre-fill with dummy values so the memory is already paged-in
	std::vector<nanoseconds> durations(conf.iterations());
	std::vector<char> buffer(conf.msg_size());
	int const tag {42};

	MPI_ASSERT(MPI_Barrier(MPI_COMM_WORLD));
	for (ssize_t cc = -static_cast<ssize_t>(conf.warmup_iterations()),
	  cc_end = static_cast<ssize_t>(conf.iterations());
	  cc < cc_end; ++cc)
	{
		// ignore warm-up runs for accounting
		// *end* of last warm-up loop is our "overall starting time"
		if (cc == 0) {
			time_start = time_end;
		}

		timed_loop(conf.busy_before());
		steady_clock::time_point time_start_mpi{steady_clock::now()};

		if (!mpi_rank)
		{
			MPI_ASSERT(MPI_Send(buffer.data(),
			                    static_cast<int>(conf.msg_size()),
			                    MPI_CHAR,
			                    !mpi_rank,
			                    tag,
			                    MPI_COMM_WORLD));

			timed_loop(conf.busy_inside());

			MPI_ASSERT(MPI_Recv(buffer.data(),
			                    0,
			                    MPI_CHAR,
			                    !mpi_rank,
			                    tag,
			                    MPI_COMM_WORLD,
			                    MPI_STATUS_IGNORE));
		} else
		{
			MPI_ASSERT(MPI_Recv(buffer.data(),
			                    static_cast<int>(conf.msg_size()),
			                    MPI_CHAR,
			                    !mpi_rank,
			                    tag,
			                    MPI_COMM_WORLD,
			                    MPI_STATUS_IGNORE));

			timed_loop(conf.busy_inside());

			MPI_ASSERT(MPI_Send(buffer.data(),
			                    0,
			                    MPI_CHAR,
			                    !mpi_rank,
			                    tag,
			                    MPI_COMM_WORLD));
		}

		time_end = steady_clock::now();

		// only record non-warm-up runs
		if (cc >= 0) {
			durations[static_cast<size_t>(cc)] = time_end - time_start_mpi;
		}

		timed_loop(conf.busy_after());
	}

	// ========================================================================//
	// output
	if (!mpi_rank)
	{
		nanoseconds dur_all{time_end - time_start};
		Statistics const in_mpi {durations};

		std::cout << "Total time for all iterations: "
		          << dur_all << std::endl
		          << "Average time per iteration: "
		          << dur_all / conf.iterations() << std::endl
		          << "Time in MPI per iteration..." << std::endl
		          << "\tMedian: " << in_mpi.median << std::endl
		          << "\tAverage: " << in_mpi.mean << std::endl
		          << "\tStandard Deviation: " << in_mpi.stddev << std::endl
		          << "\tMinimum: " << in_mpi.min << std::endl
		          << "\tMaximum: " << in_mpi.max << std::endl
		          << std::endl;

		if (! conf.outfile().empty())
		{
			if (write_timings(conf.outfile(), durations)) {
				std::cout << "Detailed results written to "
				          << std::quoted(conf.outfile()) << std::endl;
			}
			else
			{
				std::cerr << "Failed writing detailed results to file "
				          << std::quoted(conf.outfile()) << std::endl;
			}
		}
	}

	MPI_ASSERT(MPI_Finalize());
	exit(EXIT_SUCCESS);
}
