#!/usr/bin/env python3

import csv, os, sys

class Run:
	_id = 0

	def __init__(self, log_dir):
		self.log_dir = log_dir
		self.values = { "id": Run._id }
		Run._id = Run._id + 1
		print(f"{self.values['id']}: '{log_dir}'")

		self._parse_config()
		self._parse_log()
		self._parse_sysfs()
		self._parse_hdeem()
		self._parse_duration()

	def _parse_config(self):
		with open(self.log_dir + "/config", "r") as f:
			spinsPoll = spinsYield = yieldType = None

			for line in f:
				(intro, value) = line.split("=")
				value = value.strip('"\n')

				if 'SPINS_POLL' in intro: spinsPoll = int(value)
				elif 'SPINS_YIELD' in intro: spinsYield = int(value)
				elif 'BLOCK_TO' in intro: self.values["blockTimeout_ms"] = int(value)
				elif 'LIBRA_ITER' in intro: self.values["iterations"] = int(value)
				elif 'LIBRA_WARMUP' in intro: self.values["warmup"] = int(value)
				elif 'LIBRA_MSGSIZE' in intro: self.values["msgsize_B"] = int(value)
				elif 'COMM' in intro: self.values["backend"] = value
				elif 'YIELD_TYPE' in intro: yieldType = int(value)

			if spinsPoll < 0:
				self.values["mode"]="p"
			else:
				if spinsYield < 0:
					if yieldType == 1:
						self.values["mode"]="Y"
					else:
						self.values["mode"]="y"
				else:
					self.values["mode"]="b"


	def _parse_sysfs(self):
		with open(self.log_dir + "/sysfs", "r") as f:
			for line in f:
				(intro, value) = line.split(" -> ")
				type = intro.split(' ')[0].capitalize()
				id = intro.split(':')[1].rstrip("'")

				if value.startswith("(!) "):
					print("RAPL overflow")
					value = value[4:]

				value = value.strip('"\n').split(' ')[0]

				self.values["rapl" + type + id + '_mJ'] = int(value) / 1000


	def _parse_hdeem(self):
		file = self.log_dir + "/hdeem"
		if not os.path.isfile(file): return

		with open(file, "r") as f:
			parse = False
			for line in f:
				if not parse:
					if line.startswith('==== HDEEM statistics'): parse = True
					continue

				if line.startswith('==== HDEEM statistics total'): return

				parts = line.lstrip(' ').split(' ')
				intro = parts[0]
				value = parts[-1].strip('"\n')

				if 'Duration' in intro: self.values["hdeemDuration_ms"] = float(value) * 1000
				elif 'BLADE' in intro: self.values["hdeemNode_mJ"]   = float(value) * 1000
				elif 'CPU0' in intro: self.values["hdeemSocket0_mJ"] = float(value) * 1000
				elif 'CPU1' in intro: self.values["hdeemSocket1_mJ"] = float(value) * 1000
				elif 'DDR_AB' in intro: self.values["hdeemRamAB_mJ"] = float(value) * 1000
				elif 'DDR_CD' in intro: self.values["hdeemRamCD_mJ"] = float(value) * 1000
				elif 'DDR_EF' in intro: self.values["hdeemRamEF_mJ"] = float(value) * 1000
				elif 'DDR_GH' in intro: self.values["hdeemRamGH_mJ"] = float(value) * 1000

	def _parse_log(self):
		with open(self.log_dir + "/log", "r") as f:
			parse = False
			for line in f:
				if line[0] == '[': parse = True
				if not parse: continue

				if 'Host:' in line or line.startswith('Time ') or line.startswith('Detailed ') or line == '\n': continue
				(intro, value) = line.split(': ')
				value = int(value.split(' ')[0].strip('"\n'))

				if intro.startswith('[0] Busy time before'):
					self.values["busyBefore_us"] = value

				value /= 1000
				if 'Total time ' in intro: self.values["resultIterTotal_us"] = value
				elif 'Average time ' in intro: self.values["resultIterAvg_us"] = value
				elif '\tMedian' in intro: self.values["mpiMedian_us"] = value
				elif '\tAverage' in intro: self.values["mpiAvg_us"] = value
				elif '\tStandard Deviation' in intro: self.values["mpiStdev_us"] = value
				elif '\tMinimum' in intro: self.values["mpiMin_us"] = value
				elif '\tMaximum' in intro: self.values["mpiMax_us"] = value

	def _parse_duration(self):
		file = self.log_dir + "/duration"
		if not os.path.isfile(file): return

		with open(file, "r") as f:
			self.values["extTime_ms"] = float(f.readline().rstrip('\n')) * 1000


def main():
	dir = sys.argv[1]
	if not os.path.isdir(dir): raise(dir + " not a directory")

	runs = []
	for path, dirs, files in os.walk(dir):
		if ('log' not in files): continue
		#if (len(runs) > 10): exit(0)
		runs.append(Run(path))

	print("Creating summary CSV")
	headers = []
	for run in runs:
		headers = set(list(headers) + list(run.values.keys()))

	with open(dir + '/results.csv', 'w', newline='') as f:
		writer = csv.DictWriter(f, fieldnames=headers, dialect='unix')
		writer.writeheader()
		for run in runs:
			writer.writerow(run.values)

	print("Creating detailed CSV")
	with open(dir + '/detailed.csv', 'w', newline='') as f:
		writer = csv.writer(f, dialect='unix')
		writer.writerow(["id", "mpiTime_ns"])
		for run in runs:
			with open(run.log_dir + '/detailed', 'r') as df:
				for line in df:
					if line.startswith('# '): continue
					writer.writerow([run.values["id"], line.rstrip('\n')])

if __name__ == "__main__":
	main()
