#!/usr/bin/env bash

#SBATCH --time=5:00      # walltime
#SBATCH --nodes=1            # number of nodes
#SBATCH --ntasks-per-node=2  # number of processor cores per node
#SBATCH --mem=0              # we want the node's whole memory
#SBATCH --exclusive          # use node(s) exclusively
#SBATCH --comment=no_monitoring # disable PIKA (https://doc.zih.tu-dresden.de/software/pika/), only works for exclusibe jobs
#
#SBATCH --job-name="Libra"
##
#SBATCH --constrain fs_lustre_scratch2 # we need '/scratch'
#SBATCH --wait-all-nodes=1   # Do not begin execution until all nodes are ready for use
#SBATCH --cpu-freq=high
#SBATCH --partition=haswell    # 2 sockets * 12 cores (* 2 threads on *_ht partition) {Intel Xeon CPU E5-2680 v3 @ 2.50GHz}

# make sure we run in a decent shell
[[ ${BASH_VERSINFO[0]} -lt 5 ]] && exec "bash-5.1/bin/bash" "$0" "$@"
declare -r MPIRUN='openmpi-4.1.1/bin/mpirun'
declare -r LIBRA='Libra/libra'
declare -r LOG_MAIN="logs/Libra/$(date -In)"

declare ENERGY_METHOD=' '
declare -r PATH_RAPL='/sys/devices/virtual/powercap/intel-rapl'
declare -r PATH_HDEEM='/usr/bin'


declare RUN=1
declare WAIT_BEFORE=0
declare SPINS_POLL=0
declare SPINS_YIELD=0
declare BLOCK_TO=50
declare COMM='rc,sm,self'
declare LIBRA_ITER=10000
declare LIBRA_WARMUP=130
declare LIBRA_MSGSIZE=0

mkdir -p "${LOG_MAIN}"
cp "${BASH_SOURCE[0]}" "${LOG_MAIN}/run.sh"


function energy_init()
{
  [[ -d "${PATH_RAPL}" && -r "${PATH_RAPL}/intel-rapl:0/intel-rapl:0:0/energy_uj" ]] && ENERGY_METHOD+='sysfs '
  [[ -f "${PATH_HDEEM}/checkHdeem" ]] && ENERGY_METHOD+='hdeem '

  if [[ "$ENERGY_METHOD" == *' sysfs '* ]]; then
    declare -gA RAPL_START_VALUE
    local -a DOMAINS

    DOMAINS=( $(find "${PATH_RAPL}" -name name ) )  # find all domains
    DOMAINS=( ${DOMAINS[@]%/name} )  # filter for the...
    DOMAINS=( ${DOMAINS[@]#"${PATH_RAPL}/"} )     # ... actual domain name/path

    for DOMAIN in "${DOMAINS[@]}"; do
      RAPL_START_VALUE[$DOMAIN]=0 # dummy value to create key
    done
  fi
}

function energy_start()
{
  if [[ "$ENERGY_METHOD" == *' sysfs '* ]]; then
    local DOMAIN

    for DOMAIN in "${!RAPL_START_VALUE[@]}"; do
      RAPL_START_VALUE[$DOMAIN]=$(< "${PATH_RAPL}"/${DOMAIN}/energy_uj )
    done
  fi

  if [[ "$ENERGY_METHOD" == *' hdeem '* ]]; then
    "${PATH_HDEEM}/clearHdeem"
    "${PATH_HDEEM}/startHdeem"
  fi

}

function energy_stop()
{
  if [[ "$ENERGY_METHOD" = *' sysfs '* ]]; then
    local DOMAIN VALUE NAME

    for DOMAIN in "${!RAPL_START_VALUE[@]}"; do
      NAME=$(< "${PATH_RAPL}/${DOMAIN}/name" )
      VALUE=$(< "${PATH_RAPL}/${DOMAIN}/energy_uj" )
      (( VALUE -= ${RAPL_START_VALUE[$DOMAIN]} ))

      # check overflow
      [[ $VALUE -lt 0 ]] && VALUE='(!) '"$(( VALUE + $(< "${PATH_RAPL}/${DOMAIN}/max_energy_range_uj" ) ))"

      echo "${NAME%-*} '${DOMAIN##*/}' -> ${VALUE} uJ" >> "${LOG}/sysfs"
    done
  fi

  if [[ "$ENERGY_METHOD" == *' hdeem '* ]]; then
    "${PATH_HDEEM}/stopHdeem"
    "${PATH_HDEEM}/checkHdeem" &> "${LOG}/hdeem"
  fi
}


function single_run()
{
  export OMPI_MCA_plm_rsh_pass_environ_mca_params='yes'
  export OMPI_MCA_pml='ucx'
  export OMPI_MCA_osc='ucx'
  export OMPI_MCA_btl='^vader,tcp,openib,uct'


  LIBRA_ITER=10000
  LIBRA_WARMUP=$(( LIBRA_ITER / 15 )) # 6,67% of iterations as warm-up
  { YIELD_TYPE=$(< /proc/sys/kernel/sched_compat_yield); } 2>/dev/null


  local RUN_NAME="over__poll=${SPINS_POLL}_yield=${SPINS_YIELD}_block=${BLOCK_TO}_yield-type=${YIELD_TYPE}_iter=${LIBRA_ITER}_warmup=${LIBRA_WARMUP}/comm=${COMM}/msgsize=${LIBRA_MSGSIZE}/wait-before=${WAIT_BEFORE}"
  local LOG="${LOG_MAIN}/${RUN_NAME}"
  mkdir -p "${LOG}"

  [[ -n "${COMM}" ]] && COMM_ARG="-x UCX_TLS=${COMM},self"

  # we store our details file in SHM first as Taurus just sucks :-/
  local -r TMPFILE="/dev/shm/$$.${RANDOM}"

  # safe config for future reference
  declare -p ${!OMPI_*} ${!SPINS_*} BLOCK_TO ${!LIBRA_*} ${!SLURM_*} COMM YIELD_TYPE TMPFILE > "${LOG}/config"

  : "${RUN_NAME//_/ }" # expand and throw away; expansion = $_ in next line
  echo -n "${RUN}: ${_//// }"

  local -r TS=${EPOCHREALTIME}
  energy_start

  # run the benchmark
  local RESULT=$( \
   LC_ALL=C \
   "${MPIRUN}" \
   --verbose \
   -np 1 -x LIBRA_ITER=${LIBRA_ITER} -x LIBRA_WARMUP=${LIBRA_WARMUP} \
   -x LIBRA_MSGSIZE=${LIBRA_MSGSIZE} ${COMM_ARG} \
   ./over-pin.sh "${LIBRA}" -b ${WAIT_BEFORE} -o "${TMPFILE}" : \
   -np 1 -x LIBRA_ITER=${LIBRA_ITER} -x LIBRA_WARMUP=${LIBRA_WARMUP} \
   -x LIBRA_MSGSIZE=${LIBRA_MSGSIZE} -x OMPI_MCA_mpi_spins_poll=${SPINS_POLL} \
   -x OMPI_MCA_mpi_spins_yield=${SPINS_YIELD} \
   -x OMPI_MCA_mpi_block_timeout=${BLOCK_TO} ${COMM_ARG} \
   ./over-pin.sh "${LIBRA}" 2>&1 \
  )

  energy_stop
  local -r TE=${EPOCHREALTIME}
  local -r DUR=$(bc <<< "$TE - $TS")

  mv "${TMPFILE}" "${LOG}/detailed"

  # keep log + time in a file
  echo "$RESULT" > "${LOG}/log"
  echo "$DUR" > "${LOG}/duration"
  sync

  # append median runtime to progress output
  : "${RESULT##*Median: }"
  echo " -> ${_%%$'\n'*} (${DUR})"

  (( ++RUN ))
}

energy_init

# dummy run to make sure the binaries are preloaded
LC_ALL=C "${MPIRUN}" -np 1 "${LIBRA}" -h &> /dev/null
sync


LIBRA_MSGSIZE=0
WAIT_BEFORE=50
COMM=''

# poll
SPINS_POLL=-1
single_run

# yield
SPINS_POLL=0
SPINS_YIELD=-1
single_run

# block
SPINS_POLL=0
SPINS_YIELD=0
single_run
